import uuid
import adage.graph
from node import Node

def mknode(dag,task, nodename = 'node', depends_on = None):
    nodeobj = Node(str(uuid.uuid1()),nodename,task)

    dag.addNode(nodeobj)

    if not depends_on: return nodeobj

    for parent in depends_on:
        dag.addEdge(parent,nodeobj)
    return nodeobj

def mk_dag():
    return adage.graph.AdageDAG()



import networkx as nx
import dagstate
import subprocess
import nodestate
import datetime

def node_visible(node,time,start,stop):
    norm_node_time = (node.define_time-start)/(stop-start)
    return norm_node_time < time    

def colorize_graph(dag,normtime = None):
    colorized = nx.DiGraph()
    allnodes = dag.nodes()
    starttimes = [dag.getNode(n).define_time for n in allnodes]
    stoptimes  = [dag.getNode(n).ready_by_time for n in allnodes]
    start,stop = min(starttimes),max(stoptimes)

    # print "total delta: {}".format(stop-start)
    for node in dag.nodes():
        nodeobj = dag.getNode(node)
        
        time = start + normtime*(stop-start)

        # joblength = nodeobj.ready_by_time-nodeobj.submit_time
        # print 'joblength: {} {}'.format(nodeobj.name,joblength)

        color = None
        if dagstate.upstream_failure(dag,nodeobj):
            color = 'blue'
        if start <= time < nodeobj.submit_time:
            color = 'grey'
        if nodeobj.submit_time <= time < nodeobj.ready_by_time:
            color = 'yellow'
        if nodeobj.ready_by_time <= time <= stop:
            if nodeobj.state()==nodestate.FAILED:
                color = 'red'
            if nodeobj.state()==nodestate.SUCCESS:
                color = 'green'

        
        visible = node_visible(nodeobj,normtime,start,stop)
        # hmtimes = [start,nodeobj.submit_time,nodeobj.ready_by_time,stop]
        # hmnormes = [(t-start)/(stop-start) for t in hmtimes]
        # print 'times: {} {} {} {}'.format(normtime,hmnormes,color,visible)

        style = 'filled' if visible else 'invis'
        dot_attr = {'label':'{} '.format(nodeobj.name), 'style':style, 'color': color}

        colorized.add_node(node,dot_attr)
        for pre in dag.predecessors(node):
            colorized.add_edge(pre,node)
    
    # print '------------'
    dotformat = nx.drawing.nx_pydot.to_pydot(colorized)
    dotformat.set_label(datetime.datetime.fromtimestamp(time).strftime('%Y-%m-%d %H:%M:%S'))
    for e in dotformat.get_edges():
        edge_visible =  node_visible(dag.getNode(e.get_destination().replace('"','')),normtime,start,stop)
        if not edge_visible:
            e.set_style('invis')
    return dotformat

def print_dag(dag,name,trackdir,time = None):
    pngfilename = '{}/{}.png'.format(trackdir,name) 
    with open(pngfilename,'w') as pngfile:
        p = subprocess.Popen(['dot','-Tpng','-Gsize=18,12\!','-Gdpi=100'], stdout = pngfile, stdin = subprocess.PIPE)
        p.communicate(colorize_graph(dag,time).to_string())


def qualifiedname(thing):
    if thing.__module__ != '__main__':
        return '{}.{}'.format(thing.__module__,thing.__name__)
    else:
        return thing.__name__


class Functor(object):
    def __init__(self,func):
        self.qualname = qualifiedname(func)
        self.func = func
        self.args = None
        self.kwargs = None
    
    def __call__(self,dag):
        updated = self.kwargs.copy()
        updated.update(dag = dag)
        return self.func(*self.args,**updated)

    def setargs(self,*args,**kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

def functorize(func):
    def sig(*args,**kwargs):
        instance = Functor(func)
        instance.setargs(*args,**kwargs)
        return instance
    func.s = sig
    return func
    
class Rule(object):
    def __init__(self,predicate,body):
        self.predicate = predicate
        self.body = body

    def applicable(self,dag):
        return self.predicate(dag)

    def apply(self,dag):
        return self.body(dag)
    
class DelayedCallable(object):
    def __init__(self,func):
        self.qualname = qualifiedname(func)
        self.func = func
        self.args = None
        self.kwargs = None
        
    def __repr__(self):
        return '<Delayed: {}>'.format(self.func)
        
    def __call__(self):
        assert self.kwargs is not None
        assert self.args is not None
        return self.func(*self.args,**self.kwargs)

    def setargs(self,*args,**kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

def adagetask(func):
    try:
        from celery import shared_task
        func.celery = shared_task(func)
    except ImportError:
        pass        

    def sig(*args,**kwargs):
        instance = DelayedCallable(func)
        instance.setargs(*args,**kwargs)
        return instance
    func.s = sig
    return func





import time
import os
import shutil
import subprocess
import dagstate
import datetime        
import networkx as nx
import adage.visualize as viz

class GifTracker(object):
    def __init__(self,gifname,workdir,frames = 20):
        self.gifname = gifname
        self.workdir = workdir
        self.frames = frames
        
    def initialize(self,dag):
        pass
        
    def track(self,dag):
        pass
        
    def finalize(self,dag):
        if os.path.exists(self.workdir):
            shutil.rmtree(self.workdir)
        os.makedirs(self.workdir)
        for i in range(self.frames+1):
            viz.print_dag(dag,'dag_{:02}'.format(i),self.workdir,time = i/float(self.frames))
        subprocess.call('convert -delay 50 $(ls {}/*.png|sort) {}'.format(self.workdir,self.gifname),shell = True)
        shutil.rmtree(self.workdir)
                
class TextSnapShotTracker(object):
    def __init__(self,logfilename,mindelta):
        self.logfilename = logfilename
        self.mindelta = mindelta
        self.last_update = None

    def initialize(self,dag):
        if not os.path.exists(os.path.dirname(self.logfilename)):
            os.makedirs(os.path.dirname(self.logfilename))
        with open(self.logfilename,'w') as logfile:
            timenow = datetime.datetime.now().isoformat()
            logfile.write('========== ADAGE LOG BEGIN at {} ==========\n'.format(timenow))
        self.update(dag)

    def track(self,dag):
        now = time.time()
        if not self.last_update or (now-self.last_update) > self.mindelta:
            self.last_update = now
            self.update(dag)

    def update(self,dag):
        with open(self.logfilename,'a') as logfile:
            logfile.write('---------- snapshot at {}\n'.format(datetime.datetime.now().isoformat()))
            for node in nx.topological_sort(dag):
                nodeobj = dag.getNode(node)
                logfile.write('name: {} obj: {} submitted: {}\n'.format(nodeobj.name,nodeobj,nodeobj.submitted))
    def finalize(self,dag):
        with open(self.logfilename,'a') as logfile:
            self.update(dag)
            timenow = datetime.datetime.now().isoformat()
            logfile.write('========== ADAGE LOG END at {} ==========\n'.format(timenow))
        
class SimpleReportTracker(object):
    def __init__(self,log):
        self.log = log
    def initialize(self,dag):
        pass
    def track(self,dag):
        pass
    def finalize(self,dag):
        successful, failed, notrun = 0,0,0
        for node in dag.nodes():
            nodeobj = dag.getNode(node)
            if dagstate.node_status(nodeobj):
                successful+=1
            if dagstate.node_ran_and_failed(nodeobj):
                failed+=1
                self.log.error("node: {} failed. reason: {}".format(nodeobj,nodeobj.backend.fail_info(nodeobj.result)))
            if dagstate.upstream_failure(dag,nodeobj):
                notrun+=1
        self.log.info('successful: {} | failed: {} | notrun: {} | total: {}'.format(successful,failed,notrun,len(dag.nodes())))
        



import networkx as nx
import logging
import os
import dagstate
import time
from adage import trackers
from adage import nodestate
from adage.dagutils import mk_dag, mknode
from adage.decorators import functorize, adagetask, Rule


#silene pyflakes
assert mk_dag
assert mknode
assert functorize
assert adagetask
assert Rule

__all__ = ['decorators','dagutils','trackers']

log = logging.getLogger(__name__)

def validate_finished_dag(dag):
    for node in dag:
        nodeobj = dag.getNode(node)
        if nodeobj.submitted:
            sanity = all([nodeobj.submitted > dag.getNode(x).ready_by_time for x in dag.predecessors(node)])
            if not sanity:
                return False
    return True

def nodes_left_or_rule(dag,rules):
    nodes_we_could_run = [node for node in dag.nodes() if not dagstate.upstream_failure(dag,dag.getNode(node))]
    nodes_running_or_defined = [x for x in nodes_we_could_run if dagstate.node_defined_or_waiting(dag.getNode(x))]

    if any(rule.applicable(dag) for rule in rules):
        return True

    log.debug('nodes we could run: %s',nodes_we_could_run)
    if nodes_running_or_defined:
        log.debug('%s nodes that could be run or are running are left.',len(nodes_running_or_defined))
        log.debug('nodes are: %s',[dag.node[n] for n in nodes_running_or_defined])
        return True
    else:
        log.info('no nodes can be run anymore')
        return False

def update_dag(dag,rules):
    #iterate rules in reverse so we can safely pop items
    for i,rule in reversed([x for x in enumerate(rules)]):
        if rule.applicable(dag):
            log.info('extending graph.')
            rule.apply(dag)
            rules.pop(i)
        else:
            log.debug('rule not ready yet')

def process_dag(backend,dag,*unused):
    for node in nx.topological_sort(dag):
        nodeobj = dag.getNode(node)

        if not nodeobj.backend:
            nodeobj.backend = backend

        log.debug("working on node: %s with obj %s",node,nodeobj)

        if nodeobj.submitted:
            log.debug("node already submitted. continue")
            continue;
        if dagstate.upstream_ok(dag,nodeobj):
            log.info('submitting %s job',nodeobj)
            nodeobj.result = backend.submit(nodeobj.task)
            submit_time = time.time()
            nodeobj.submitted = submit_time

        if dagstate.upstream_failure(dag,nodeobj):
            log.warning('not submitting node: %s due to upstream failure',node)
    
def rundag(dag,rules, track = False, backend = None, loggername = None, workdir = None, trackevery = 1):
    if loggername:
        global log
        log = logging.getLogger(loggername)
    
    ## funny behavior of multiprocessing Pools means that
    ## we can not have backendsubmit = multiprocsetup(2)    in the function sig
    ## so we only initialize them here
    if not backend:
        from backends import MultiProcBackend
        backend = MultiProcBackend(2)

    if not workdir:
        workdir = os.getcwd()

    trackerlist = [trackers.SimpleReportTracker(log)]
    
    if track:
        trackerlist += [trackers.GifTracker(gifname = '{}/workflow.gif'.format(workdir), workdir = '{}/track'.format(workdir))]
        trackerlist += [trackers.TextSnapShotTracker(logfilename = '{}/adagesnap.txt'.format(workdir), mindelta = trackevery)]
        
    for t in trackerlist: t.initialize(dag)
    #while we have nodes that can be submitted

    try:
      while nodes_left_or_rule(dag,rules):
          update_dag(dag,rules)
          process_dag(backend,dag,rules)
          for t in trackerlist: t.track(dag)
          time.sleep(1)
    except:
      log.error('some weird exception caught in adage process loop')
      raise  

    log.info('all running jobs are finished.')
    
    for node in dag.nodes():
        #check node status one last time so we pick up the finishing times
        dagstate.node_status(dag.getNode(node))

    log.info('track last time')
        
    for t in trackerlist: t.finalize(dag)

    log.info('validating execution')

    if not validate_finished_dag(dag):
        log.error('DAG execution not validating')
        raise RuntimeError('DAG execution not validating')
    log.info('execution valid. (in terms of execution order)')
    
    if any(dag.getNode(x).state() == nodestate.FAILED for x in dag.nodes()):
        log.error('raising RunTimeError due to failed jobs')
        raise RuntimeError('DAG execution failed')



import nodestate
import time


class Node(object):
    def __init__(self,identifier,name,task):
        self.identifier = identifier
        self.name = name
        self.submit_time = None
        self.ready_by_time = None
        self.the_state = nodestate.DEFINED
        self.define_time = time.time()        
        self.task = task
        self.the_result = None
        self.the_backend = None

    def __repr__(self):
        return '<Node name: {} id: {} state: {}>'.format(self.name,self.identifier,self.state())

    def update_state(self):
        #if we do not have a result object
        #that means it's not submitted yet
        if not self.the_result:
            self.the_state = nodestate.DEFINED
            return
        
        #if we have a resultobject
        #but the result is not ready
        #the node is still running
        if not self.ready():
            self.the_state = nodestate.RUNNING
            return
            
        #if it's ready it's either successful
        #or failed
        if self.successful():
            self.the_state = nodestate.SUCCESS
            return
        else:
            self.the_state = nodestate.FAILED
            return

    def state(self):
        self.update_state()
        return self.the_state

    @property
    def backend(self):
        return self.the_backend
    
    @backend.setter
    def backend(self,backend):
        self.the_backend = backend
    
    @property
    def submitted(self):
        return self.submit_time
        
    @submitted.setter
    def submitted(self,time):
        self.the_state = nodestate.RUNNING
        self.submit_time = time
        
    @property
    def result(self):
        return self.the_result
        
    @result.setter
    def result(self,result):
        self.the_result = result
    
    def ready(self):
        if not self.the_result:
            return False

        ready = self.backend.ready(self.the_result)
        if not ready:
            return False
        if not self.ready_by_time:
            self.ready_by_time = time.time()
        return True

    def result_of(self):
        return self.backend.result_of(self.the_result)
        
    def successful(self):
        if not self.the_result:
            return False
        return self.backend.successful(self.the_result)



import networkx as nx
class AdageDAG(nx.DiGraph):
    def addNode(self,nodeobj):
        self.add_node(nodeobj.identifier, {'nodeobj': nodeobj})
    def addEdge(self,fromobj,toobj):
        self.add_edge(fromobj.identifier,toobj.identifier)
    def getNode(self,ident):
        return self.node[ident]['nodeobj']
    def getNodeByName(self,name):
        matching = [x for x in self.nodes() if self.getNode(x).name == name]
        return self.getNode(matching[0]) if (len(matching) == 1) else None
        


import multiprocessing
import sys
import traceback

class MultiProcBackend(object):
    def __init__(self,poolsize):
        self.pool = multiprocessing.Pool(poolsize)
    def submit(self,task):
        return self.pool.apply_async(task)
    def result_of(self,result):
        return result.get()
    def ready(self,result):
        return result.ready()
    def successful(self,result):
        if not self.ready(result): return False
        return result.successful()
    def fail_info(self,result):
        try:
            self.result_of(result)
        except:
            t,v,tb =    sys.exc_info()
            traceback.print_tb(tb)
            return (t,v)
            
class CeleryBackend(object):
    def __init__(self,app):
        self.app = app
    def submit(self,task):
        self.app.set_current()
        return task.func.celery.apply_async(task.args,task.kwargs,throw = False)
    def result_of(self,result):
        return result.get()
    def ready(self,result):
        return result.ready()
    def successful(self,result):
        return result.successful()
    def fail_info(self,result):
        try:
            self.result_of(result)
        except:
            return sys.exc_info()
        
class DummyResult(object):        
    pass
    
class DummyBackend(object):
    def submit(self,task):
        if task:
            pass
        return DummyResult()
    def result_of(self,*unused):
        return None
    def ready(self,*unused):
        return True
    def successful(self,*unused):
        return False
    def fail_info(self,*unused):
        return 'cannot give reason :( '
